"""This sub-module offers methods to automatically retrieve the graphs from KGOBO repository."""

from .mod import MOD
from .fbbt import FBBT
from .bto import BTO
from .chmo import CHMO
from .mco import MCO
from .opmi import OPMI
from .fbdv import FBDV
from .mpath import MPATH
from .spd import SPD
from .vt import VT
from .ehdaa2 import EHDAA2
from .wbls import WBLS
from .rxno import RXNO
from .omp import OMP
from .ero import ERO
from .gno import GNO
from .xco import XCO
from .amphx import AMPHX
from .clyh import CLYH
from .ncro import NCRO
from .iao import IAO
from .exo import EXO
from .swo import SWO
from .obcs import OBCS
from .symp import SYMP
from .taxrank import TAXRANK
from .apo import APO
from .clo import CLO
from .cmo import CMO
from .hso import HSO
from .obi import OBI
from .cheminf import CHEMINF
from .mp import MP
from .labo import LABO
from .olatdv import OLATDV
from .chebi import CHEBI
from .aeo import AEO
from .geno import GENO
from .sbo import SBO
from .uo import UO
from .mop import MOP
from .ogms import OGMS
from .ncbitaxon import NCBITAXON
from .pw import PW
from .fovt import FOVT
from .xpo import XPO
from .zfs import ZFS
from .rs import RS
from .omo import OMO
from .fix import FIX
from .vto import VTO
from .uberon import UBERON
from .bfo import BFO
from .htn import HTN
from .aism import AISM
from .wbbt import WBBT
from .hao import HAO
from .so import SO
from .mondo import MONDO
from .ddpheno import DDPHENO
from .idomal import IDOMAL
from .fbcv import FBCV
from .trans import TRANS
from .scdo import SCDO
from .dron import DRON
from .rbo import RBO
from .ncit import NCIT
from .fma import FMA
from .rex import REX
from .cob import COB
from .sibo import SIBO
from .pdro import PDRO
from .ogg import OGG
from .hancestro import HANCESTRO
from .go import GO
from .plana import PLANA
from .oae import OAE
from .mmusdv import MMUSDV
from .hsapdv import HSAPDV
from .miro import MIRO
from .emapa import EMAPA
from .gecko import GECKO
from .tads import TADS
from .fao import FAO
from .ecao import ECAO
from .opl import OPL
from .tgma import TGMA
from .bco import BCO
from .ico import ICO
from .zeco import ZECO
from .pdumdv import PDUMDV
from .aro import ARO
from .oarcs import OARCS
from .planp import PLANP
from .doid import DOID
from .omrse import OMRSE
from .ppo import PPO
from .ovae import OVAE
from .zp import ZP
from .stato import STATO
from .ecto import ECTO
from .xao import XAO
from .miapa import MIAPA
from .mi import MI
from .ecocore import ECOCORE
from .mmo import MMO
from .eupath import EUPATH
from .obib import OBIB
from .ido import IDO
from .tto import TTO
from .pr import PR
from .nbo import NBO
from .wbphenotype import WBPHENOTYPE
from .cio import CIO
from .clao import CLAO
from .upa import UPA
from .zfa import ZFA
from .ma import MA
from .cdno import CDNO
from .ohd import OHD
from .agro import AGRO
from .txpo import TXPO
from .pato import PATO
from .hom import HOM
from .eco import ECO
from .ddanat import DDANAT
from .bspo import BSPO
from .mro import MRO
from .pco import PCO
from .ornaseq import ORNASEQ
from .hp import HP
from .dpo import DPO
from .cl import CL

__all__ = [
	"MOD", "FBBT", "BTO", "CHMO", "MCO", "OPMI", "FBDV", "MPATH", "SPD", "VT",
	"EHDAA2", "WBLS", "RXNO", "OMP", "ERO", "GNO", "XCO", "AMPHX", "CLYH",
	"NCRO", "IAO", "EXO", "SWO", "OBCS", "SYMP", "TAXRANK", "APO", "CLO", "CMO",
	"HSO", "OBI", "CHEMINF", "MP", "LABO", "OLATDV", "CHEBI", "AEO", "GENO",
	"SBO", "UO", "MOP", "OGMS", "NCBITAXON", "PW", "FOVT", "XPO", "ZFS", "RS",
	"OMO", "FIX", "VTO", "UBERON", "BFO", "HTN", "AISM", "WBBT", "HAO", "SO",
	"MONDO", "DDPHENO", "IDOMAL", "FBCV", "TRANS", "SCDO", "DRON", "RBO", "NCIT",
	"FMA", "REX", "COB", "SIBO", "PDRO", "OGG", "HANCESTRO", "GO", "PLANA",
	"OAE", "MMUSDV", "HSAPDV", "MIRO", "EMAPA", "GECKO", "TADS", "FAO", "ECAO",
	"OPL", "TGMA", "BCO", "ICO", "ZECO", "PDUMDV", "ARO", "OARCS", "PLANP",
	"DOID", "OMRSE", "PPO", "OVAE", "ZP", "STATO", "ECTO", "XAO", "MIAPA",
	"MI", "ECOCORE", "MMO", "EUPATH", "OBIB", "IDO", "TTO", "PR", "NBO", "WBPHENOTYPE",
	"CIO", "CLAO", "UPA", "ZFA", "MA", "CDNO", "OHD", "AGRO", "TXPO", "PATO",
	"HOM", "ECO", "DDANAT", "BSPO", "MRO", "PCO", "ORNASEQ", "HP", "DPO", "CL",
]